#include <cmath>
#include <iostream>
#include <sstream>
#include <sys/time.h>
#include <vector>

#define GRN "\33[92m"
#define YEL "\33[93m"
#define RES "\33[0m"

template<typename T>
std::string toStr(const T& nr)
{
  std::ostringstream os;
  os << nr;
  return os.str();
}

template<typename T>
T strTo(const std::string &s)
{
  T num;
  std::istringstream is(s);
  is >> num;
  return num;
}

template<typename T>
T strTo(const char* s)
{
  T num;
  std::istringstream is(s);
  is >> num;
  return num;
}

class TimeCPU
{
  private:
    struct timeval ts;
    double t;

  public:
    TimeCPU()
    : t(0.0)
    {}

    void start()
    {
      gettimeofday(&ts, NULL);
    }

    double read()
    {
      struct timeval te;
      gettimeofday(&te, NULL);
      t = (double)(te.tv_sec-ts.tv_sec)+0.000001*(double)(te.tv_usec-ts.tv_usec);
      return t;
    }

    const double& time() const { return t; }
    double time() { return t; }
};

inline std::ostream& operator<<(std::ostream& os, const TimeCPU& t)
{
  return os << t.time();
}

inline double operator/(const double& l, const TimeCPU& r)
{
  return l/r.time();
}

inline double operator/(const TimeCPU& l, const double& r)
{
  return l.time()/r;
}

inline double operator*(const double& l, const TimeCPU& r)
{
  return l*r.time();
}

inline double operator*(const TimeCPU& l, const double& r)
{
  return l.time()*r;
}

inline double operator+(const double& l, const TimeCPU& r)
{
  return l+r.time();
}

inline double operator+(const TimeCPU& l, const double& r)
{
  return l.time()+r;
}

inline double operator-(const double& l, const TimeCPU& r)
{
  return l-r.time();
}

inline double operator-(const TimeCPU& l, const double& r)
{
  return l.time()-r;
}

void init(std::vector< std::vector<double> >& A,
          std::vector< std::vector<double> >& B,
          std::vector< std::vector<double> >& C,
          int N, int M, int L)
{
  A.resize(N);
  B.resize(L);
  C.resize(N);
  for(int i=0; i<N; ++i) A[i].resize(L);
  for(int i=0; i<L; ++i) B[i].resize(M);
  for(int i=0; i<N; ++i) C[i].resize(M);

  for(int i=0; i<N; ++i)
  {
    for(int j=0; j<L; ++j)
    {
      A[i][j] = (i+j)%11+5.0;
    }
  }
  
  for(int i=0; i<L; ++i)
  {
    for(int j=0; j<M; ++j)
    {
      B[i][j] = (i+j)%36+5.0;
    }
  }

  for(int i=0; i<N; ++i)
  {
    for(int j=0; j<M; ++j)
    {
      C[i][j] = 0.0;
    }
  }
}

bool validate(std::vector< std::vector<double> >& A,
              std::vector< std::vector<double> >& B,
              std::vector< std::vector<double> >& C,
              int N, int M, int L)
{
  std::vector< std::vector<double> > D;
  D.resize(N);
  for(int i=0; i<N; ++i) D[i].resize(M);

  for(int i=0; i<N; ++i)
  {
    for(int j=0; j<M; ++j)
    {
      D[i][j] = 0.0;
    }
  }

  for(int i=0; i<N; ++i)
  {
    for(int j=0; j<M; ++j)
    {
      for(int k=0; k<L; ++k)
      {
        D[i][j] += A[i][k] * B[k][j];
      }
    }
  }

  bool error=false;
  for(int i=0; i<N; ++i)
  {
    for(int j=0; j<M; ++j)
    {
      if( fabs(D[i][j] - C[i][j]) > 0.001 )
      {
        error = true;
        i=N; j=M;
      }
    }
  }

  return !error;
}


//start to modify

/*void matMul(std::vector< std::vector<double> >& A,
            std::vector< std::vector<double> >& B,
            std::vector< std::vector<double> >& C,
            int N, int M, int L)
{
  for(int i=0; i<N; ++i)
  {
    for(int j=0; j<M; ++j)
    {
      for(int k=0; k<L; ++k)
      {
        C[i][j] += A[i][k] * B[k][j];
      }
    }
  }
}*/

/*
basic version:
compiled: g++ -o lab05 lab05.cpp -W -Wall -pedantic -pipe
performance: 20.4976 s

compiled: g++ -o lab05 lab05.cpp -W -Wall -pedantic -pipe -O3
performance: 7.3412 s
*/

/*
void matMul(std::vector< std::vector<double> >& A,
            std::vector< std::vector<double> >& B,
            std::vector< std::vector<double> >& C,
            int N, int M, int L)
{
  for(int i=0; i<N; ++i)
  {
    for(int k=0; k<L; ++k) //loop exchange
    {
      for(int j=0; j<M; ++j) //loop exchange
      {
        C[i][j] += A[i][k] * B[k][j];
      }
    }
  }
}

/*
optimization above: loop exchange
performance: 0.415183 s
*/


void matMul(const std::vector< std::vector<double> >& A, //const
            const std::vector< std::vector<double> >& B, //const
            std::vector< std::vector<double> >& C,
            const int N, const int M, const int L) //const
{
  for(int i=0; i<N; ++i)
  {
    for(int k=0; k<L; ++k)
    {
      const double Aik = A[i][k]; //memory access reduction
      for(int j=0; j<M; j+=4) //loop unrolling
      {
        C[i][j]   += Aik * B[k][j];
        C[i][j+1] += Aik * B[k][j+1];
        C[i][j+2] += Aik * B[k][j+2];
        C[i][j+3] += Aik * B[k][j+3];
      }
    }
  }
}

/*
optimization above: const variables, loop unrolling, memory access reduction
performance: 0.397561 s
*/

//stop to modify

int main()
{
  std::cout << GRN << "-----=====#####=====-----\n" << RES;
  TimeCPU tCPU;

  int N=1024;
  int M=N;
  int L=N;
  std::vector< std::vector<double> > A, B, C;

  init(A, B, C, N, M, L);

  tCPU.start();
  matMul(A, B, C, N, M, L);
  tCPU.read();

  const double Gflop=2.0*N*M*L*0.000000001;
  const double perf=Gflop/tCPU;

  std::cout << "Result";
  if( validate(A, B, C, N, M, L) )
    std::cout << GRN << " - correct\n" << RES;
  else
    std::cout << YEL << " - error\n" << RES;
  std::cout << "The matMul() function executed in " << tCPU << " [s]\n";
  std::cout << "Executed " << Gflop << " billion operations\n";
  std::cout << "Efficiency " << YEL << perf << RES << " [Gflop/s]\n";
  std::cout << "Size: " << N << 'x' << M << 'x' << L << '\n';
  std::cout << GRN << "-----=====#####=====-----\n" << RES;
  return 0;
}
